/*
  Project Name: Player
  Description: This program is the player panel. Here, the player can log in with their RFID chip. 
  Then, the program automatically pairs with the master via ESP-NOW. After that, the player can
  receive messages from the master, which are then displayed on the OLED screen. The variable 
  BOARD_ID must be set differently for each panel. Addresses 6 - 124 are available.
  Author: Lena Hagenauer
  Date: 05 January 2025
  Version: 2.0
  License: MIT License
*/

// Include libraries
// Status LED
#include <Adafruit_NeoPixel.h>
#define LED_PIN   2
#define LED_COUNT 1
Adafruit_NeoPixel LED(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);

// RFID
#include <SPI.h>
#include <MFRC522.h>

#define SS_PIN  15  
#define RST_PIN 0

MFRC522 mfrc522(SS_PIN, RST_PIN);

// Name Information
char name[16] = " ";

// OLED
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

#define OLED_RESET     -1 
#define SCREEN_ADDRESS 0x3C
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// WIFI & ESPNOW
#include <ESP8266WiFi.h>
#include <espnow.h>

uint8_t channel = 1;


#define MAX_CHANNEL 13  // 11 in North America or 13 in Europe

unsigned long currentMillis = millis(); 
unsigned long lastTime = 0;  
unsigned long timerDelay = 2000;  // send readings timer

uint8_t broadcastAddressX[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

uint8_t clientMacAddress[6];

enum PairingStatus {PAIR_REQUEST, PAIR_REQUESTED, PAIR_PAIRED, };
PairingStatus pairingStatus = PAIR_REQUEST;

enum MessageType {PAIRING, DATA,};
MessageType messageType;


// Updates DHT readings every 10 seconds
// const long interval = 10000; 
unsigned long previousMillis = 0;    // will store last time DHT was updated 

// Structure to send data
typedef struct struct_message {
  uint8_t msgType;
  uint8_t id;
  char message[168];
} struct_message;

// Structure for pairing message
typedef struct struct_pairing {       
    uint8_t msgType;
    uint8_t id;
    uint8_t macAddr[6];
    uint8_t channel;
    char name[64];
} struct_pairing;

// Variables definitions
struct_message myData;
struct_message incomingReadings;
struct_pairing pairingData;

// Massage for display
char message[168] = "";
int state = 0;
bool newMessage = false;


#define BOARD_ID 6 // need to be change for evry player panel
unsigned long start;

// Callback when data is sent
void OnDataSent(uint8_t *mac_addr, uint8_t sendStatus) {
  Serial.print("Last Packet Send Status: ");
  if (sendStatus == 0){
    Serial.println("Delivery success");
  }
  else{
    Serial.println("Delivery fail");
  }
}

// Function to print MAC address
void printMAC(const uint8_t * mac_addr){
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  Serial.print(macStr);
}

// Callback when data is received
void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) {
  Serial.print("Size of message : ");
  Serial.print(len);
  Serial.print(" from ");
  printMAC(mac);
  Serial.println();
  uint8_t type = incomingData[0];
  switch (type) {
  case DATA :  
    memcpy(&incomingReadings, incomingData, sizeof(incomingReadings));
    Serial.print(len);
    Serial.print(" Data bytes received from: ");
    printMAC(mac);
    Serial.println();
    Serial.println(incomingReadings.message);
    
    strcpy(message, incomingReadings.message);
    newMessage = true;

    break;

  case PAIRING:
    memcpy(&pairingData, incomingData, sizeof(pairingData));
    if (pairingData.id == 0) {                // the message comes from server
      Serial.print("Pairing done for ");
      printMAC(pairingData.macAddr);
      Serial.print(" on channel " );
      Serial.print(pairingData.channel);    // channel used by the server
      Serial.print(" in ");
      Serial.print(millis()-start);
      Serial.println("ms");
      //esp_now_del_peer(pairingData.macAddr);
      //esp_now_del_peer(mac);
      esp_now_add_peer(pairingData.macAddr, ESP_NOW_ROLE_COMBO, pairingData.channel, NULL, 0); // add the server to the peer list 
      pairingStatus = PAIR_PAIRED ;            // set the pairing status
    }
    break;
  }  
}

// Function to read the MAC address
void readGetMacAddress(){
  String val = WiFi.macAddress();
  Serial.println(val);
  char* endPtr; 
  clientMacAddress[0] = strtol(val.c_str(), &endPtr, 16); // read the first starting at the beginning of the buffer. this initializes endPtr as a pointer to the ':' after the first number 
  for (int i = 1;  (*endPtr) && (i < 6); i++) {
    clientMacAddress[i] = strtol(endPtr + 1, &endPtr, 16); // using +1 for the pointer as we want to skip the ':'
  }

  for (int i = 0; i < 6; i++) {
    Serial.print(clientMacAddress[i], HEX);
    if (i != 5) Serial.print(F(":"));
  }
}

// Function for autoPairing
PairingStatus autoPairing(){
  switch(pairingStatus) {
  case PAIR_REQUEST:
    Serial.print("Pairing request on channel "  );
    Serial.println(channel);
  
    // clean esp now
    esp_now_deinit();
    WiFi.mode(WIFI_STA);
    // set WiFi channel   
    wifi_promiscuous_enable(1);
    wifi_set_channel(channel);
    wifi_promiscuous_enable(0);
    //WiFi.printDiag(Serial);
    WiFi.disconnect();

    // Init ESP-NOW
    if (esp_now_init() != 0) {
      Serial.println("Error initializing ESP-NOW");
    }
    esp_now_set_self_role(ESP_NOW_ROLE_COMBO);
    // set callback routines
    esp_now_register_send_cb(OnDataSent);
    esp_now_register_recv_cb(OnDataRecv);
    
    // set pairing data to send to the server
    pairingData.id = BOARD_ID;     
    pairingData.channel = channel;
    pairingData.macAddr[0] = clientMacAddress[0];
    pairingData.macAddr[1] = clientMacAddress[1];
    pairingData.macAddr[2] = clientMacAddress[2];
    pairingData.macAddr[3] = clientMacAddress[3];
    pairingData.macAddr[4] = clientMacAddress[4];
    pairingData.macAddr[5] = clientMacAddress[5];

    strcpy(pairingData.name, name);
    Serial.println(name);
 
    previousMillis = millis();
    // add peer and send request
    Serial.println(esp_now_send(broadcastAddressX, (uint8_t *) &pairingData, sizeof(pairingData)));
    pairingStatus = PAIR_REQUESTED;
    break;

  case PAIR_REQUESTED:
    // time out to allow receiving response from server
    currentMillis = millis();
    if(currentMillis - previousMillis > 1000) {
      previousMillis = currentMillis;
      // time out expired,  try next channel
      channel ++;
      if (channel > MAX_CHANNEL) {
        channel = 0;
      }
      pairingStatus = PAIR_REQUEST; 
    }
    break;

  case PAIR_PAIRED:
    //Serial.println("Paired!");
    break;
  }
  return pairingStatus;
} 


//draw dots and D&D
void drawStart(char* str){
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0,0); 
  display.println("  D&D");
  display.setTextSize(1);
  display.println();
  display.println("warten auf Spieler");
  display.setTextSize(3);
  display.println(str);
  display.display();

}

// Function returns True when RFID Name is read
boolean waitForPlayer(){
  readDataRFID();
  if(strcmp(name, " ") != 0){
    writeText(name,2);
    return true;
  }else{
   return false;
  }

}

// Function to write text on the OLED display
void writeText(char* str, int size){

  display.clearDisplay();
  display.setTextSize(size);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0,0); 
  display.write(str);
  display.display();

}

// Funtion to read RFID chip
void readDataRFID(){
  
  if ( ! mfrc522.PICC_IsNewCardPresent()){
    return;
  }
  // Select one of the cards
  if ( ! mfrc522.PICC_ReadCardSerial()){
  return;
  }

  byte buffer[18]; 
  byte size = sizeof(buffer);
  uint8_t pageAddr = 0x07;// depending on RFID chip

  mfrc522.MIFARE_Read(pageAddr, buffer, &size);

  for (int i = 0; i < 16; i++) {
    if(buffer[i+2] != 47){
      name[i] = buffer[i+2];
      Serial.println();
      //Serial.write(buffer[i+1]);
    }else{
     i=18;     
    }
  }
  mfrc522.PICC_HaltA();
}

void setup() {
  // Init Serial Monitor
  Serial.begin(74880);


  // Status LED
  LED.begin();
  LED.clear();


  // RFID 
  SPI.begin();
  mfrc522.PCD_Init(); 


  // OLED
   // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;); // Don't proceed, loop forever
  }
  display.clearDisplay();
 
  // Set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);
  readGetMacAddress();
  //Serial.println(WiFi.macAddress());
  WiFi.disconnect();

  // Init ESP-NOW
  if (esp_now_init() != 0) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Set ESP-NOW Role
  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);
    
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  esp_now_register_send_cb(OnDataSent);

  pairingData.id = 2;
  
  LED.setPixelColor(0, LED.Color(0, 150, 150));
  LED.show();

  while(!waitForPlayer()){

    drawStart(" ");
    delay(250);
    drawStart(".");
    delay(500);
    drawStart("..");
    delay(250);
    drawStart("...");
    delay(250);
    drawStart("....");
    delay(250);
    drawStart(".....");
    delay(250);
    drawStart("......");
    delay(250);
    drawStart(".......");
    delay(250);
  }


}

// Function to indicate a green OK signal with the LED
void greendOK(){
  for(int i = 0; i<3; i++){
    LED.setBrightness(150);
    LED.setPixelColor(0, LED.Color(0, 255, 0));
    LED.show();
    delay(250);
    LED.clear();
    LED.show();
    delay(250);
    LED.clear();

  }

}

// Function to create a blue pulsing effect with the LED
void bluePuls(){

   for(int i = 200; i > 50; i--){
      LED.setBrightness(i);
      LED.setPixelColor(0, LED.Color(0, 0, 255));
      LED.show();
      delay(2);
    }
   for(int i = 50; i < 200; i++){
      LED.setBrightness(i);
      LED.setPixelColor(0, LED.Color(0, 0, 255));
      LED.show();
      delay(2);
   }
}

// Function to indicate a message received signal with the LED
void getMessage(){
  for(int i = 0; i<3; i++){
    LED.setBrightness(150);
    LED.setPixelColor(0, LED.Color(255, 255, 0));
    LED.show();
    delay(250);
    LED.clear();
    LED.show();
    delay(250);
    LED.clear();
  }
}
 

void loop() { 

  if(state == 0){
    bluePuls();
    if(autoPairing() == PAIR_PAIRED ){
      writeText("Connected ",1);
      state = 1;
      greendOK();
    }
  }
  if(newMessage){
    writeText(message,1);
    newMessage = false;
    getMessage();
  }
  
}